@@ -1,3 +1,4 @@
+use crate::check_var::{check_vars, check_vars_in_rewriter};
 use crate::fixer::{Fixer, FixerError, SerializableFixer};
 use crate::rule::referent_rule::RuleRegistration;
 use crate::rule::Rule;
@@ -66,7 +67,7 @@ impl SerializableRuleCore {
     if let Some(utils) = &self.utils {
       let env = env
         .register_local_utils(utils)
-        .map_err(RuleConfigError::Rule)?;
+        .map_err(RuleConfigError::Utils)?;
       Ok(env)
     } else {
       Ok(env)
@@ -77,14 +78,14 @@ impl SerializableRuleCore {
     &self,
     env: &DeserializeEnv<L>,
   ) -> RResult<HashMap<String, Rule<L>>> {
-    let mut matchers = HashMap::new();
-    let Some(constraints) = &self.constraints else {
-      return Ok(matchers);
+    let mut constraints = HashMap::new();
+    let Some(serde_cons) = &self.constraints else {
+      return Ok(constraints);
     };
-    for (key, ser) in constraints {
-      matchers.insert(key.to_string(), env.deserialize_rule(ser.clone())?);
+    for (key, ser) in serde_cons {
+      constraints.insert(key.to_string(), env.deserialize_rule(ser.clone())?);
     }
-    Ok(matchers)
+    Ok(constraints)
   }
 
   fn get_fixer<L: Language>(&self, env: &DeserializeEnv<L>) -> RResult<Option<Fixer<L>>> {
@@ -98,86 +99,47 @@ impl SerializableRuleCore {
 
   fn get_matcher_from_env<L: Language>(&self, env: &DeserializeEnv<L>) -> RResult<RuleCore<L>> {
     let rule = env.deserialize_rule(self.rule.clone())?;
-    let matchers = self.get_constraints(env)?;
-    let vars = Self::check_var_in_constraints(&rule, &matchers)?;
-    let vars = Self::check_var_in_transform(vars, &self.transform)?;
+    let constraints = self.get_constraints(env)?;
     let transform = self.transform.clone();
     let fixer = self.get_fixer(env)?;
-    Self::check_var_in_fix(vars, &fixer)?;
     Ok(
       RuleCore::new(rule)
-        .with_matchers(matchers)
+        .with_matchers(constraints)
         .with_utils(env.registration.clone())
         .with_transform(transform)
         .with_fixer(fixer),
     )
   }
 
-  fn check_var_in_constraints<'r, L: Language>(
-    rule: &'r Rule<L>,
-    constraints: &'r HashMap<String, Rule<L>>,
-  ) -> RResult<HashSet<&'r str>> {
-    let mut vars = rule.defined_vars();
-    for rule in constraints.values() {
-      for var in rule.defined_vars() {
-        vars.insert(var);
-      }
-    }
-    for var in constraints.keys() {
-      let var: &str = var;
-      if !vars.contains(var) {
-        return Err(RuleConfigError::UndefinedMetaVar(
-          var.to_owned(),
-          "constraints",
-        ));
-      }
-    }
-    Ok(vars)
-  }
-
-  fn check_var_in_transform<'r>(
-    mut vars: HashSet<&'r str>,
-    transform: &'r Option<HashMap<String, Transformation>>,
-  ) -> RResult<HashSet<&'r str>> {
-    let Some(transform) = transform else {
-      return Ok(vars);
-    };
-    for var in transform.keys() {
-      vars.insert(var);
-    }
-    for trans in transform.values() {
-      let needed = trans.used_vars();
-      if !vars.contains(needed) {
-        return Err(RuleConfigError::UndefinedMetaVar(
-          needed.to_string(),
-          "transform",
-        ));
-      }
-    }
-    Ok(vars)
-  }
-
-  fn check_var_in_fix<L: Language>(vars: HashSet<&str>, fixer: &Option<Fixer<L>>) -> RResult<()> {
-    let Some(fixer) = fixer else {
-      return Ok(());
-    };
-    for var in fixer.used_vars() {
-      if !vars.contains(&var) {
-        return Err(RuleConfigError::UndefinedMetaVar(var.to_string(), "fix"));
-      }
-    }
-    Ok(())
-  }
-
   pub fn get_matcher<L: Language>(&self, env: DeserializeEnv<L>) -> RResult<RuleCore<L>> {
     let env = self.get_deserialize_env(env)?;
-    self.get_matcher_from_env(&env)
+    let ret = self.get_matcher_from_env(&env)?;
+    check_vars(&ret.rule, &ret.constraints, &ret.transform, &ret.fixer)?;
+    Ok(ret)
+  }
+
+  /// this is needed for checking metavar defined in containing vars
+  pub(crate) fn get_rewriter<L: Language>(
+    &self,
+    env: DeserializeEnv<L>,
+    upper_vars: &HashSet<&str>,
+  ) -> RResult<RuleCore<L>> {
+    let env = self.get_deserialize_env(env)?;
+    let ret = self.get_matcher_from_env(&env)?;
+    check_vars_in_rewriter(
+      &ret.rule,
+      &ret.constraints,
+      &ret.transform,
+      &ret.fixer,
+      upper_vars,
+    )?;
+    Ok(ret)
   }
 }
 
 pub struct RuleCore<L: Language> {
   rule: Rule<L>,
-  matchers: HashMap<String, Rule<L>>,
+  constraints: HashMap<String, Rule<L>>,
   kinds: Option<BitSet>,
   pub(crate) transform: Option<HashMap<String, Transformation>>,
   pub fixer: Option<Fixer<L>>,
@@ -197,8 +159,11 @@ impl<L: Language> RuleCore<L> {
   }
 
   #[inline]
-  pub fn with_matchers(self, matchers: HashMap<String, Rule<L>>) -> Self {
-    Self { matchers, ..self }
+  pub fn with_matchers(self, constraints: HashMap<String, Rule<L>>) -> Self {
+    Self {
+      constraints,
+      ..self
+    }
   }
 
   #[inline]
@@ -225,7 +190,10 @@ impl<L: Language> RuleCore<L> {
 
   pub fn defined_vars(&self) -> HashSet<&str> {
     let mut ret = self.rule.defined_vars();
-    for constraint in self.matchers.values() {
+    for v in self.utils.get_local_util_vars() {
+      ret.insert(v);
+    }
+    for constraint in self.constraints.values() {
       for var in constraint.defined_vars() {
         ret.insert(var);
       }
@@ -237,6 +205,35 @@ impl<L: Language> RuleCore<L> {
     }
     ret
   }
+
+  pub(crate) fn do_match<'tree, D: Doc<Lang = L>>(
+    &self,
+    node: Node<'tree, D>,
+    env: &mut Cow<MetaVarEnv<'tree, D>>,
+    enclosing_env: Option<&MetaVarEnv<'tree, D>>,
+  ) -> Option<Node<'tree, D>> {
+    if let Some(kinds) = &self.kinds {
+      if !kinds.contains(node.kind_id().into()) {
+        return None;
+      }
+    }
+    let ret = self.rule.match_node_with_env(node, env)?;
+    if !env.to_mut().match_constraints(&self.constraints) {
+      return None;
+    }
+    if let Some(trans) = &self.transform {
+      let lang = ret.lang();
+      let rewriters = self.utils.get_rewriters();
+      let env = env.to_mut();
+      if let Some(enclosing) = enclosing_env {
+        apply_env_transform(trans, lang, env, rewriters, enclosing);
+      } else {
+        let enclosing = env.clone();
+        apply_env_transform(trans, lang, env, rewriters, &enclosing);
+      };
+    }
+    Some(ret)
+  }
 }
 impl<L: Language> Deref for RuleCore<L> {
   type Target = Rule<L>;
@@ -250,7 +247,7 @@ impl<L: Language> Default for RuleCore<L> {
   fn default() -> Self {
     Self {
       rule: Rule::default(),
-      matchers: HashMap::default(),
+      constraints: HashMap::default(),
       kinds: None,
       transform: None,
       fixer: None,
@@ -265,21 +262,7 @@ impl<L: Language> Matcher<L> for RuleCore<L> {
     node: Node<'tree, D>,
     env: &mut Cow<MetaVarEnv<'tree, D>>,
   ) -> Option<Node<'tree, D>> {
-    if let Some(kinds) = &self.kinds {
-      if !kinds.contains(node.kind_id().into()) {
-        return None;
-      }
-    }
-    let ret = self.rule.match_node_with_env(node, env)?;
-    if !env.to_mut().match_constraints(&self.matchers) {
-      return None;
-    }
-    if let Some(trans) = &self.transform {
-      let lang = ret.lang();
-      let rewriters = self.utils.get_rewriters();
-      apply_env_transform(trans, lang, env.to_mut(), rewriters);
-    }
-    Some(ret)
+    self.do_match(node, env, None)
   }
 
   fn potential_kinds(&self) -> Option<BitSet> {
@@ -315,13 +298,13 @@ mod test {
 
   #[test]
   fn test_rule_with_constraints() {
-    let mut matchers = HashMap::new();
-    matchers.insert(
+    let mut constraints = HashMap::new();
+    constraints.insert(
       "A".to_string(),
       Rule::Regex(RegexMatcher::try_new("a").unwrap()),
     );
     let rule =
-      RuleCore::new(Rule::Pattern(Pattern::new("$A", TypeScript::Tsx))).with_matchers(matchers);
+      RuleCore::new(Rule::Pattern(Pattern::new("$A", TypeScript::Tsx))).with_matchers(constraints);
     let grep = TypeScript::Tsx.ast_grep("a");
     assert!(grep.root().find(&rule).is_some());
     let grep = TypeScript::Tsx.ast_grep("bbb");
@@ -405,78 +388,4 @@ transform:
     let transformed = env.get_transformed("C").expect("should transform");
     assert_eq!(String::from_utf8_lossy(transformed), "a");
   }
-
-  #[test]
-  fn test_defined_vars() {
-    let env = DeserializeEnv::new(TypeScript::Tsx);
-    let ser_rule: SerializableRuleCore = from_str(
-      r"
-rule: {pattern: $A = $B}
-constraints:
-  A: { pattern: $C = $D }
-transform:
-  E:
-    substring:
-      source: $B
-      startCar: 1",
-    )
-    .expect("should deser");
-    let matcher = ser_rule.get_matcher(env).expect("should parse");
-    assert_eq!(
-      matcher.defined_vars(),
-      ["A", "B", "C", "D", "E"].into_iter().collect()
-    );
-  }
-
-  fn get_undefined(src: &str) -> (String, &str) {
-    let env = DeserializeEnv::new(TypeScript::Tsx);
-    let ser_rule: SerializableRuleCore = from_str(src).expect("should deser");
-    match ser_rule.get_matcher(env) {
-      Err(RuleConfigError::UndefinedMetaVar(name, section)) => (name, section),
-      _ => panic!("should error"),
-    }
-  }
-
-  #[test]
-  fn test_undefined_vars_in_constraints() {
-    let (name, section) = get_undefined(
-      r"
-rule: {pattern: $A}
-constraints: {B: {pattern: bbb}}
-",
-    );
-    assert_eq!(name, "B");
-    assert_eq!(section, "constraints");
-  }
-  #[test]
-  fn test_undefined_vars_in_transform() {
-    let (name, section) = get_undefined(
-      r"
-rule: {pattern: $A}
-constraints: {A: {pattern: $C}}
-transform:
-  B:
-    replace: {source: $C, replace: a, by: b }
-  D:
-    replace: {source: $E, replace: a, by: b }
-",
-    );
-    assert_eq!(name, "E");
-    assert_eq!(section, "transform");
-  }
-  #[test]
-  fn test_undefined_vars_in_fix() {
-    let (name, section) = get_undefined(
-      r"
-rule: {pattern: $A}
-constraints: {A: {pattern: $C}}
-transform:
-  B:
-    replace: {source: $C, replace: a, by: b }
-fix: $D
-",
-    );
-    assert_eq!(name, "D");
-    assert_eq!(section, "fix");
-  }
 }
