@@ -3,7 +3,7 @@ use crate::rule_core::RuleCore;
 
 use ast_grep_core::meta_var::MetaVariable;
 use ast_grep_core::source::{Content, Edit};
-use ast_grep_core::{Doc, Language, Matcher, Node};
+use ast_grep_core::{Doc, Language, Node, NodeMatch};
 use schemars::JsonSchema;
 use serde::{Deserialize, Serialize};
 
@@ -47,7 +47,7 @@ impl Rewrite {
       .iter()
       .filter_map(|id| rewriters.get(id)) // TODO: better handling
       .collect();
-    let edits = find_and_make_edits(nodes, &rules);
+    let edits = find_and_make_edits(nodes, &rules, ctx);
     let rewritten = if let Some(joiner) = &self.join_by {
       let mut ret = vec![];
       let mut edits = edits.into_iter();
@@ -77,28 +77,34 @@ impl Rewrite {
 }
 
 type Bytes<D> = [<<D as Doc>::Source as Content>::Underlying];
-fn find_and_make_edits<D: Doc>(
-  nodes: Vec<Node<D>>,
+fn find_and_make_edits<'n, D: Doc>(
+  nodes: Vec<Node<'n, D>>,
   rules: &[&RuleCore<D::Lang>],
+  ctx: &Ctx<'_, 'n, D>,
 ) -> Vec<Edit<D::Source>> {
   nodes
     .into_iter()
-    .flat_map(|n| replace_one(n, rules))
+    .flat_map(|n| replace_one(n, rules, ctx))
     .collect()
 }
 
-fn replace_one<D: Doc>(node: Node<D>, rules: &[&RuleCore<D::Lang>]) -> Vec<Edit<D::Source>> {
+fn replace_one<'n, D: Doc>(
+  node: Node<'n, D>,
+  rules: &[&RuleCore<D::Lang>],
+  ctx: &Ctx<'_, 'n, D>,
+) -> Vec<Edit<D::Source>> {
   let mut edits = vec![];
   for child in node.dfs() {
     for rule in rules {
-      // TODO: should it inherit meta_var_env?
-      // if yes, it is very confusing recursive env is not inherited
+      let mut env = std::borrow::Cow::Borrowed(ctx.enclosing_env);
+      // NOTE: we inherit meta_var_env from enclosing rule
+      // but match env will NOT inherited recursively!
       // e.g. $B is matched in parent linter and it is inherited.
-      // $C is matched in rewriter and is not inherited in recursive rewriter?
-      // if no, it is not flexible enough to support cross-rule communication
-      // but it is clear that no env inhertance in rewriter.
+      // $C is matched in rewriter but is NOT inherited in recursive rewriter
+      // this is to enable recursive rewriter to match sub nodes
       // in future, we can use the explict `expose` to control env inheritance
-      if let Some(nm) = rule.match_node(child.clone()) {
+      if let Some(n) = rule.do_match(child.clone(), &mut env, Some(ctx.enclosing_env)) {
+        let nm = NodeMatch::new(n, env.into_owned());
         edits.push(nm.make_edit(rule, rule.fixer.as_ref().expect("TODO")));
         // stop at first fix, skip duplicate fix
         break;
@@ -138,6 +144,7 @@ mod test {
   use crate::rule_core::SerializableRuleCore;
   use crate::test::TypeScript;
   use crate::GlobalRules;
+  use std::collections::HashSet;
 
   fn apply_transformation(
     rewrite: Rewrite,
@@ -145,33 +152,26 @@ mod test {
     pat: &str,
     rewriters: GlobalRules<TypeScript>,
   ) -> String {
-    let grep = TypeScript::Tsx.ast_grep(src);
-    let root = grep.root();
-    let mut nm = root.find(pat).expect("should find");
-    let before_vars: Vec<_> = nm.get_env().get_matched_variables().collect();
-    let mut ctx = Ctx {
-      lang: &TypeScript::Tsx,
-      transforms: &Default::default(),
-      env: nm.get_env_mut(),
-      rewriters,
-    };
-    let after_vars: Vec<_> = ctx.env.get_matched_variables().collect();
-    assert_eq!(
-      before_vars, after_vars,
-      "rewrite should not write back to env"
-    );
-    rewrite.compute(&mut ctx).expect("should have transforms")
+    compute_rewritten(src, pat, rewrite, rewriters).expect("should have transforms")
   }
+
   macro_rules! str_vec {
     ( $($a: expr),* ) => { vec![ $($a.to_string()),* ] };
   }
 
-  fn make_rewriter(pairs: &[(&str, &str)]) -> GlobalRules<TypeScript> {
+  fn make_rewriters(pairs: &[(&str, &str)]) -> GlobalRules<TypeScript> {
+    make_rewriters_with_rewriter(pairs, Default::default())
+  }
+
+  fn make_rewriters_with_rewriter(
+    pairs: &[(&str, &str)],
+    vars: HashSet<&str>,
+  ) -> GlobalRules<TypeScript> {
     let rewriters = GlobalRules::default();
     for (key, ser) in pairs {
       let serialized: SerializableRuleCore = from_str(ser).unwrap();
       let env = DeserializeEnv::new(TypeScript::Tsx).with_rewriters(&rewriters);
-      let rule = serialized.get_matcher(env).unwrap();
+      let rule = serialized.get_rewriter(env, &vars).unwrap();
       rewriters.insert(key, rule).unwrap();
     }
     rewriters
@@ -184,7 +184,7 @@ mod test {
       rewriters: str_vec!["rewrite"],
       join_by: None,
     };
-    let rewriters = make_rewriter(&[("rewrite", "{rule: {kind: number}, fix: '810'}")]);
+    let rewriters = make_rewriters(&[("rewrite", "{rule: {kind: number}, fix: '810'}")]);
     let ret = apply_transformation(rewrite, "log(t(1, 2, 3))", "log($A)", rewriters);
     assert_eq!(ret, "t(810, 810, 810)");
   }
@@ -196,7 +196,7 @@ mod test {
       rewriters: str_vec!["re1", "re2"],
       join_by: None,
     };
-    let rewriters = make_rewriter(&[
+    let rewriters = make_rewriters(&[
       ("re1", "{rule: {regex: '^1$'}, fix: '810'}"),
       ("re2", "{rule: {regex: '^2$'}, fix: '1919'}"),
     ]);
@@ -211,7 +211,7 @@ mod test {
       rewriters: str_vec!["re1"],
       join_by: None,
     };
-    let rewriters = make_rewriter(&[
+    let rewriters = make_rewriters(&[
       ("ignored", "{rule: {regex: '^2$'}, fix: '1919'}"),
       ("re1", "{rule: {kind: number}, fix: '810'}"),
     ]);
@@ -227,7 +227,7 @@ mod test {
       join_by: None,
     };
     // first match wins the rewrite
-    let rewriters = make_rewriter(&[
+    let rewriters = make_rewriters(&[
       ("re2", "{rule: {regex: '^2$'}, fix: '1919'}"),
       ("re1", "{rule: {kind: number}, fix: '810'}"),
     ]);
@@ -243,7 +243,7 @@ mod test {
       join_by: None,
     };
     // parent node wins fix, even if rule comes later
-    let rewriters = make_rewriter(&[
+    let rewriters = make_rewriters(&[
       ("re1", "{rule: {kind: number}, fix: '810'}"),
       ("re2", "{rule: {kind: array}, fix: '1919'}"),
     ]);
@@ -258,7 +258,7 @@ mod test {
       rewriters: str_vec!["re1"],
       join_by: Some(" + ".into()),
     };
-    let rewriters = make_rewriter(&[("re1", "{rule: {kind: number}, fix: '810'}")]);
+    let rewriters = make_rewriters(&[("re1", "{rule: {kind: number}, fix: '810'}")]);
     let ret = apply_transformation(rewrite, "log(t(1, 2, 3))", "log($A)", rewriters);
     assert_eq!(ret, "810 + 810 + 810");
   }
@@ -279,23 +279,26 @@ transform:
       rewriters: [re1]
 fix: $D
     "#;
-    let rewriters = make_rewriter(&[("re1", rule)]);
+    let rewriters = make_rewriters(&[("re1", rule)]);
     let ret = apply_transformation(rewrite, "[1, [2, [3, [4]]]]", "$A", rewriters);
     assert_eq!(ret, "1, 2, 3, 4");
   }
 
   #[test]
-  fn test_should_not_inherit_match_env() {
+  fn test_should_inherit_match_env() {
     let rewrite = Rewrite {
       source: "$A".into(),
       rewriters: str_vec!["re"],
       join_by: None,
     };
-    let rewriters = make_rewriter(&[("re", "{rule: {pattern: $B}, fix: '123'}")]);
+    let rewriters = make_rewriters(&[("re", "{rule: {pattern: $C}, fix: '123'}")]);
     let ret = apply_transformation(rewrite.clone(), "[1, 2]", "[$A, $B]", rewriters.clone());
     assert_eq!(ret, "123");
-    let ret = apply_transformation(rewrite, "[1, 1]", "[$A, $B]", rewriters);
+    let ret = apply_transformation(rewrite.clone(), "[1, 1]", "[$A, $C]", rewriters.clone());
     assert_eq!(ret, "123");
+    // should not match $C so no rewrite
+    let ret = apply_transformation(rewrite, "[1, 2]", "[$A, $C]", rewriters);
+    assert_eq!(ret, "1");
   }
 
   #[test]
@@ -305,17 +308,50 @@ fix: $D
       rewriters: str_vec!["re"],
       join_by: None,
     };
-    let rewriters = make_rewriter(&[("re", "{rule: {pattern: $B}, fix: '123'}")]);
-    let grep = TypeScript::Tsx.ast_grep("[1, 2]");
+    let rewriters = make_rewriters(&[("re", "{rule: {pattern: $B}, fix: '123'}")]);
+    let ret = compute_rewritten("[1, 2]", "[$B, $C]", rewrite, rewriters);
+    assert_eq!(ret, None);
+  }
+
+  #[test]
+  fn test_rewrite_use_enclosing_env() {
+    let rewrite = Rewrite {
+      source: "$A".into(),
+      rewriters: str_vec!["re"],
+      join_by: None,
+    };
+    let mut vars = HashSet::new();
+    vars.insert("C");
+    let rewriters =
+      make_rewriters_with_rewriter(&[("re", "{rule: {pattern: $B}, fix: '$B == $C'}")], vars);
+    let ret = apply_transformation(rewrite, "[1, 2]", "[$A, $C]", rewriters);
+    assert_eq!(ret, "1 == 2");
+  }
+
+  fn compute_rewritten(
+    src: &str,
+    pat: &str,
+    rewrite: Rewrite,
+    rewriters: GlobalRules<TypeScript>,
+  ) -> Option<String> {
+    let grep = TypeScript::Tsx.ast_grep(src);
     let root = grep.root();
-    let mut nm = root.find("[$B, $C]").expect("should find");
+    let mut nm = root.find(pat).expect("should find");
+    let before_vars: Vec<_> = nm.get_env().get_matched_variables().collect();
+    let env = nm.get_env_mut();
+    let enclosing = env.clone();
     let mut ctx = Ctx {
       lang: &TypeScript::Tsx,
       transforms: &Default::default(),
-      env: nm.get_env_mut(),
+      env,
       rewriters,
+      enclosing_env: &enclosing,
     };
-    let ret = rewrite.compute(&mut ctx);
-    assert_eq!(ret, None);
+    let after_vars: Vec<_> = ctx.env.get_matched_variables().collect();
+    assert_eq!(
+      before_vars, after_vars,
+      "rewrite should not write back to env"
+    );
+    rewrite.compute(&mut ctx)
   }
 }
